from core.core30_context.context_dependency_graph import context_dependencies, context_producer
from core.core30_context.context import Context

from ..model.remote_sstic_context import RemoteSSTICContext
from ..interact import target_choice  # trigger context dependency

import requests
import socket
import json


def last_context_for_step(step):
    ctxt = RemoteSSTICContext.query \
        .filter_by(step=step) \
        .order_by(RemoteSSTICContext.created_at.desc()) \
        .limit(1) \
        .one_or_none()
    return ctxt


def check_remote_context(ctxt: RemoteSSTICContext) -> bool:
    if not ctxt:
        return False

    burp0_url = f"{ctxt.status_url}"
    burp0_headers = {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:124.0) Gecko/20100101 Firefox/124.0",
        "Accept": "*/*", "Accept-Language": "fr,fr-FR;q=0.8,en-US;q=0.5,en;q=0.3", "Connection": "close"
    }
    out = requests.get(burp0_url, headers=burp0_headers).text
    out = json.loads(out)
    if 'expire' not in out:
        print(f"[-] Session {ctxt.instance_id} is expired")
        return False

    ports_to_confirm = [ctxt.port1, *([ctxt.port2] if ctxt.port2 < 0 else [])]
    for port in ports_to_confirm:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(5)
        try:
            sock.connect((ctxt.ip, port))
            sock.close()
        except Exception as e:
            print(f"[-] Not able to connect to {ctxt.ip}:{port} ({e})")
            return False
    return True


def new_context(step, secret_url) -> RemoteSSTICContext:
    burp0_url = f"{secret_url}/{step if step != 'step5' else 'step4'}/start"
    burp0_headers = {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:124.0) Gecko/20100101 Firefox/124.0",
        "Accept": "*/*", "Accept-Language": "fr,fr-FR;q=0.8,en-US;q=0.5,en;q=0.3", "Connection": "close"}
    out = json.loads(requests.post(burp0_url, headers=burp0_headers).text)
    if 'ports' not in out:
        print(f"[-] Bad answer received: {json.dumps(out)}")
        raise Exception(out)
    if len(out['ports']) == 1:
        ctxt = RemoteSSTICContext.create(step=step,
                                         ip='163.172.99.233',
                                         port1=out['ports'][0][0],
                                         instance_id=out['nonce'])
    elif len(out['ports']) == 2:
        ctxt = RemoteSSTICContext.create(step=step,
                                         ip='163.172.99.233',
                                         port1=out['ports'][0][0],
                                         port2=out['ports'][1][0],
                                         instance_id=out['nonce'])
    else:
        raise Exception(f"Received ports {out['ports']}, not handled (not 1 nor 2 ports returned)")
    return ctxt


@context_dependencies(('.targetctxt.step', str), ('.targetctxt.step_secret_url', str))
def get_working_context_dyn(ctxt: Context, additional_check=None):
    step = ctxt['targetctxt']['step']
    remote_ctxt = last_context_for_step(step)
    if not check_remote_context(remote_ctxt) or (additional_check and not additional_check(remote_ctxt)):
        secret_url = ctxt['targetctxt']['step_secret_url']
        try:
            stop_instance(remote_ctxt)
        except:
            pass
        remote_ctxt = new_context(step, secret_url)
        print(f"[+] New context: {remote_ctxt}")
    return remote_ctxt

@context_producer(('.targetctxt.valid_remote_context', RemoteSSTICContext),)
@context_dependencies(('.targetctxt.step', str), ('.targetctxt.step_secret_url', str))
def get_working_context(ctxt: Context, additional_check=None):
    remote_ctxt = get_working_context(additional_check)
    ctxt['targetctxt']['valid_remote_context'] = remote_ctxt
    return remote_ctxt


@context_dependencies(('.targetctxt.step', str), ('.targetctxt.step_secret_url', str))
def stop_instance(ctxt: Context, remote_context: RemoteSSTICContext):
    burp0_url = f"{ctxt['targetctxt']['step_secret_url']}/{ctxt['targetctxt']['step']}/" \
                f"{remote_context.instance_id}/stop"
    burp0_headers = {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:124.0) Gecko/20100101 Firefox/124.0",
        "Accept": "*/*", "Accept-Language": "fr,fr-FR;q=0.8,en-US;q=0.5,en;q=0.3", "Connection": "close"}
    return json.loads(requests.post(burp0_url, headers=burp0_headers).text)

