from core.core30_context.context import Context
from core.core30_context.context_dependency_graph import context_producer, context_dependencies
from ..model.local_sstic_context import LocalSSTICContext

import random
import string
import socket


def last_context_for_step(step):
    ctxt = LocalSSTICContext.query \
        .filter_by(step=step) \
        .order_by(LocalSSTICContext.created_at.desc()) \
        .limit(1) \
        .one_or_none()
    return ctxt


def check_local_context(ctxt: LocalSSTICContext) -> bool:
    ports_to_confirm = [ctxt.port]
    for port in ports_to_confirm:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(5)
        try:
            sock.connect((ctxt.ip, port))
            sock.close()
        except Exception as e:
            print(f"[-] Not able to connect to {ctxt.ip}:{port} ({e})")
            return False
    return True


def new_context(step, ip, port) -> LocalSSTICContext:
    instance_id = 'LOCAL_'+''.join([random.choice(string.ascii_lowercase) for _ in range(6)])
    ctxt = LocalSSTICContext.create(step=step,
                                    ip=ip,
                                    port=port,
                                    instance_id=instance_id)
    return ctxt


@context_dependencies(('.targetctxt.step', str),
                      ('.targetctxt.local_ip_step', str), ('.targetctxt.local_port_step', int))
def get_working_context_dyn(ctxt: Context, additional_check=None):
    step = ctxt['targetctxt']['step']
    local_ctxt = last_context_for_step(step)
    if not local_ctxt or not check_local_context(local_ctxt) or (additional_check and not additional_check(local_ctxt)):
        local_ctxt = new_context(step, ctxt['targetctxt']['local_ip_step'], ctxt['targetctxt']['local_port_step'])
        print(f"[+] New context: {local_ctxt}")
    return local_ctxt


@context_producer(('.targetctxt.valid_local_context', LocalSSTICContext))
@context_dependencies(('.targetctxt.step', str),
                      ('.targetctxt.local_ip_step', str), ('.targetctxt.local_port_step', int))
def get_working_context(ctxt: Context, additional_check=None):
    local_ctxt = get_working_context_dyn(additional_check)
    ctxt['targetctxt']['valid_local_context'] = local_ctxt
    return local_ctxt

