from core.core11_config.config import register_config_default, Config, config_dependencies
from core.core30_context.context_dependency_graph import context_producer, context_dependencies
from core.core30_context.context import Context

from typing import Callable
import enum


class ChooseStepPolicy(enum.Enum):
    CONFIG = enum.auto()
    ASK = enum.auto()

register_config_default('.sstic.choose_step', ChooseStepPolicy, ChooseStepPolicy.CONFIG)


@config_dependencies(('.sstic.current_step', str))
def choose_step_config(config: Config):
    return config['sstic']['current_step']


@context_producer(('.targetctxt.step', str), ('.targetctxt.step_secret_url', str),
                  ('.targetctxt.local_ip_step', str), ('.targetctxt.local_port_step', int))
@context_dependencies(('.interactor.ask', Callable[[...], str]))
@config_dependencies(('.sstic.choose_step', ChooseStepPolicy))
def choose_sstic_step(config: Config, ctxt: Context, chosen_step: str | None = None):
    # TODO: add the option to switch to config mode (so rewrite config)
    if not chosen_step:
        match config['sstic']['choose_step']:
            case ChooseStepPolicy.CONFIG:
                chosen_step = choose_step_config()
            case ChooseStepPolicy.ASK:
                chosen_step = ctxt['interactor']['ask']('Please give the current SSTIC step to set context to')
            case _:
                raise NotImplementedError

    ctxt.setdefault('targetctxt', {})['step'] = chosen_step
    @config_dependencies((f".targetctxt.{chosen_step}.step_secret_url", str))
    def get_step_secret_url(config: Config):
        return config['targetctxt'][chosen_step]['step_secret_url']

    @config_dependencies((f".targetctxt.{chosen_step}.local_ip", str), (f".targetctxt.{chosen_step}.local_port", int))
    def get_local_setup(config: Config):
        return config['targetctxt'][chosen_step]['local_ip'], config['targetctxt'][chosen_step]['local_port']

    ctxt['targetctxt']['step_secret_url'] = get_step_secret_url()
    ctxt['targetctxt']['local_ip_step'], ctxt['targetctxt']['local_port_step'] = get_local_setup()

    return {
        'step': ctxt['targetctxt']['step'],
        'url': ctxt['targetctxt']['step_secret_url'],
        'local_ip': ctxt['targetctxt']['local_ip_step'],
        'local_port': ctxt['targetctxt']['local_port_step']
    }
