from contextlib import contextmanager
from threading import Thread
import socket


@contextmanager
def temporary_server(host, port, on_recv, chunk_receive=1024, timeout=10):
    print(f"[.] Listening temporary server on {host}:{port}")
    server_socket = socket.socket()
    server_socket.settimeout(timeout)
    server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    server_socket.bind((host, port))
    server_socket.listen(1)

    def recv_one_and_exit():
        try:
            client_socket, client_address = server_socket.accept()
            client_socket.settimeout(timeout)
            remote_ip, _ = client_socket.getpeername()
            data = client_socket.recv(chunk_receive)
            on_recv(data, remote_ip)
        except Exception as e:
            print(f"[-] Unable to receive data: {e}")
        finally:
            try:
                client_socket.close()
            except:
                pass

    t = Thread(target=recv_one_and_exit)
    t.start()
    try:
        yield t
    finally:
        print(f"[.] Stop listening temporary server on {host}:{port}")
        server_socket.close()
        t.join()
