from core.core00_core_model.mixin.base_mixin import BaseMixinsNoJoin
from core.core00_core_model.concept.timed import utcnow

from sqlalchemy import Column, Integer, DateTime, String
from sqlalchemy.orm import Mapped, mapped_column
from datetime import datetime


class RemoteSSTICContext(*BaseMixinsNoJoin):

    __tablename__ = 'remote_sstic_target_context'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    instance_id: Mapped[str] = mapped_column(String, nullable=False)
    ip: Mapped[str] = mapped_column(String, nullable=True)
    port1: Mapped[int] = Column(Integer)
    port2: Mapped[int] = Column(Integer, default=-1)
    step: Mapped[str] = Column(String, nullable=False)

    created_at: Mapped[datetime] = Column(DateTime(timezone=True), server_default=utcnow())


    @property
    def url1(self):
        return f"http://{self.ip}:{self.port1}"

    @property
    def url2(self):
        return f"http://{self.ip}:{self.port2}"

    @property
    def status_url(self):
        return f"http://{self.ip}:8080/status/{self.instance_id}"


    def __repr__(self):
        return f"[.] {self.step} - Context {self.ip}:{self.port1}|{self.port2} " \
               f"(instance ID {self.instance_id}) created at {self.created_at}"
