from core.core05_persistent_model.policy.session import get_session
from core.core30_context.context_dependency_graph import context_dependencies
from core.core30_context.context import Context

from action.model.local_exec_context import last_working_local_context

from wussticcommon.model.reverse_port_forward import get_reverse_port_forward_manager
from wussticcommon.model.remote_sstic_context import RemoteSSTICContext
import wussticcommon.ensure_service.instance_lifecycle  # to provide valid_remote_context

import asyncio
import random
import string


@context_dependencies(('.targetctxt.valid_remote_context', RemoteSSTICContext))
def extract_context(ctxt: Context):
    return ctxt['targetctxt']['valid_remote_context'].ip, \
           ctxt['targetctxt']['valid_remote_context'].port1, ctxt['targetctxt']['valid_remote_context'].port2, \
           ctxt['targetctxt']['valid_remote_context'].instance_id

with get_session():
    SERVER_HOST, SETUP_PORT, SERVER_PORT, INSTANCE_ID = extract_context()


rpfm = get_reverse_port_forward_manager()
PUBLIC_TUNNEL_HOST = rpfm.rpf.remote_ip
PUBLIC_TUNNEL_PORT = rpfm.rpf.remote_port
LOCAL_MITM_HOST = rpfm.rpf.local_ip
LOCAL_MITM_PORT = rpfm.rpf.local_port

local_context = last_working_local_context()
additional_random = ''.join([random.choice(string.ascii_lowercase) for _ in range(6)])  # one full random attr per exec


class InteractHandler:

    def __init__(self, on_recv=None):
        self.on_recv = on_recv

    async def client_handler(self, clientside_reader, clientside_writer):
        print(f"New connection to {SERVER_HOST}, {SERVER_PORT}")
        serverside_reader, serverside_writer = await asyncio.open_connection(SERVER_HOST, SERVER_PORT)

        while True:
            try:
                if not (client_to_server := (await clientside_reader.read(1024))):
                    print("Connection closed by client")
                    client_to_server.close()
                    return
            except:
                serverside_writer.close()
                await serverside_writer.wait_closed()
                clientside_writer.close()
                await clientside_writer.wait_closed()
                return

            print(f"On socket {clientside_reader._transport._sock_fd}")
            full_context = f"{local_context.context_id}:{INSTANCE_ID}:" \
                           f"{clientside_reader._transport._sock_fd}:{additional_random}"
            print("Client: " + client_to_server.hex())

            if self.on_recv:
                server_to_client = await self.on_recv(full_context, client_to_server,
                                                      serverside_writer, serverside_reader,
                                                      clientside_writer, clientside_reader)

                if not server_to_client:
                    print("No data received from user recv")
                    return

                #print("From user: " + server_to_client.hex())
                clientside_writer.write(server_to_client)
                await clientside_writer.drain()
            else:
                serverside_writer.write(client_to_server)
                await serverside_writer.drain()

                if not (server_to_client := (await serverside_reader.read(1024))):
                    print("Connection closed by server")
                    return

                print("Server: " + server_to_client.hex())
                clientside_writer.write(server_to_client)
                await clientside_writer.drain()


    async def print_setup_read(self, reader, writer):
        if not (setup_msg := (await reader.read(1024))):
            print("\033[31mConnection closed during setup phase\033[0m")
            writer.close()
            await writer.wait_closed()
            exit()
        print(setup_msg.decode())

    # Start an instance of the challenge
    async def setup(self):
        setup_reader, setup_writer = await asyncio.open_connection(SERVER_HOST, SETUP_PORT)
        await self.print_setup_read(setup_reader, setup_writer)
        setup_writer.write(f"{PUBLIC_TUNNEL_HOST}:{PUBLIC_TUNNEL_PORT}".encode())
        await setup_writer.drain()
        await self.print_setup_read(setup_reader, setup_writer)

    # Start the MITM listener
    async def mitm_listener(self):
        server = await asyncio.start_server(self.client_handler, LOCAL_MITM_HOST, LOCAL_MITM_PORT)
        async with server:
            await server.serve_forever()

    async def main_loop(self):
        # Run the listener and the setup tasks in parallel to avoid the challenge telling us it can't connect to our listener
        with rpfm.with_ssh_forward():
            await asyncio.gather(self.mitm_listener(), self.setup())
