from .utils import send_pdu, LICENSE_KEY, pdu_handlers
from .protocol import Cmd, commands, PDU, FacingDirection

from construct import ConstructError
from typing import Container
import threading
import select
import socket
import time


class Loop:

    def __init__(self, handlers=None, username=None):
        self.background = True

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.token = None

        self.pdu_handlers = {
            k: v if handlers is None or k not in handlers else handlers[k] for k, v in pdu_handlers.items()
        }
        self.pdu_handlers.update({
            Cmd.AUTHENTICATION_RESPONSE_PDU: self.handle_token
        })

        self.messages = []
        self.message_counter = 0
        self.queue = []
        self.context = {'loop': self}

        self.main_thread = threading.Thread(target=self.receive_loop)
        self.main_thread.start()

        self.get_token(username)

        self.ping_thread = threading.Thread(target=self.ping)
        self.ping_thread.start()

        self.send_thread = threading.Thread(target=self.send_loop)
        self.send_thread.start()

    def end(self):
        print("[.] Ending")
        self.background = False
        self.sock.close()

    def heartbeat(self):
        if self.token:
            send_pdu(self.sock, Cmd.HEARTBEAT_PDU, {}, self.token)

    def ping(self):
        while self.background:
            self.heartbeat()
            time.sleep(1)

    def get_token(self, username=None):
        body = {
            "license_key": LICENSE_KEY,
            "username": b'\x44' * 14 + b'\x65\x00' if not username else username,
        }
        send_pdu(self.sock, Cmd.AUTHENTICATION_PDU, body)

    def handle_token(self, pdu, _):
        self.token = pdu['token']
        self.player_id = pdu['player_id']
        self.team = pdu['team']
        self.context.update({'player_id': pdu['player_id'], 'team': pdu['team']})
        print(f"[+] OK token: {self.token}")

    def launch_attack(self, victim_id, damage):
        self.queue.append((Cmd.CLIENT_ATTACK_PDU, {
            'victim_player_id': victim_id,
            'attack_damage': damage,
        }))

    def seize_object(self, object_id):
        self.queue.append((Cmd.OBJECT_SEIZING_PDU, {
            'object_id': object_id
        }))

    def set_position(self, facing: FacingDirection, x, y):
        cur_player = self.context['all_players'][self.context['player_id']]
        self.queue.append((Cmd.CLIENT_PLAYER_INFO_PDU, {
            'name': cur_player.name,
            'facing': facing,
            'map_id': cur_player.map_id,
            'x': x,
            'y': y,
            'is_afk': 0,
            'shielded': 0,
            'greenshard_hp': 0,
            'last_attacker': b'\x44'*13+b'\x00'*3
        }))

    def wait_for(self, ctxt_attribute, value=None):
        while not self.context or self.context.get(ctxt_attribute, None) is None or \
                (value is not None and self.context.get(ctxt_attribute, None) != value):
            time.sleep(0.1)
        return self.context[ctxt_attribute]

    def disconnect(self):
        self.queue.append((Cmd.CLIENT_DISCONNECT_PDU, {}))

    def add_to_queue(self, pdu, body):
        self.queue.append((pdu, body))

    def send_loop(self):
        while not self.token:
            time.sleep(0.3)
        print("[+] Token ok, will be able to send")
        while self.background:
            while self.queue and self.background:
                pdu, body = self.queue.pop(0)
                print(f"[+] Sending {pdu}")
                send_pdu(self.sock, pdu, body, self.token)
                time.sleep(0.2)
            time.sleep(0.3)

    def receive_loop(self):
        try:
            while self.background:
                ready_to_read, _, _ = select.select([self.sock], [], [], 0.1)
                if self.sock in ready_to_read:
                    data, addr = self.sock.recvfrom(1024)
                    try:
                        pdu: Container = PDU.parse(data)
                    except ConstructError as e:
                        continue

                    if pdu.cmdId not in self.pdu_handlers:
                        print("[-] Unknown or unsupported command, packet ignored")
                        print(pdu)
                        continue

                    try:
                        body: Container = commands[pdu.cmdId].parse(bytes(pdu.body))
                    except ConstructError:
                        print("[-] Malformed body, packet ignored")
                        continue

                    self.pdu_handlers[pdu.cmdId](body, self.context)
                    self.messages.append(body)
                    # print(self.messages[-1])
        except Exception as e:
            print(f"Received exc: {e}")
        finally:
            self.end()
