from utils import nextW, SHA256_K, print_hex, format_state_to_hash, output_to_abcdefgh, solve, PORT, HOST
from model_double import compute_one_H

from pwn import remote
from z3 import BitVec
import struct
import base64


base_msg = """0
@mafiaTEST
OnlyRequiredToSignAndSendMessages
@nothing
QQo="""


def do_for(i, j, N):
    outs = []
    for _ in range(N):
        r = remote(HOST, PORT)
        r.sendline(base_msg)
        r.sendline(str(2 * i).encode())
        r.sendline(str(j).encode())
        r.recvuntil(b"\n")
        data = r.recvuntil(b"\n").replace(b'\n', b'')
        print(f"Received for {i}, {j}: {data.decode()}")
        out = base64.b64decode(data).hex()
        outs.append(out)
        r.close()
    return outs


def solve_corpus(for_each_fault_location, constraints):
    int_arrays = for_each_fault_location
    begin = min(int_arrays.keys())
    targets = []
    faults = []
    eqs = []

    Wvars = [f"WX{i}" for i in range(64)]
    WState = [BitVec(Wvars[i], 32) if Wvars[i] not in constraints else constraints[Wvars[i]] for i in range(64)]

    Hvars = [f"H{i}" for i in range(8)]
    InitialHState = [BitVec(Hvars[i], 32) if Hvars[i] not in constraints else constraints[Hvars[i]] for i in range(8)]
    for t in int_arrays:
        for r in int_arrays[t]:
            faults.append(BitVec(f"F{len(faults)}", 32))
            targets.append(r)
            eqs.append(compute_one_H(begin, 32, t, SHA256_K, WState, faults[-1], InitialHState)[-1])

    baseline = do_for(1337, 1337, 1)[0]
    targets.append(output_to_abcdefgh(baseline))
    eqs.append(compute_one_H(begin, 32, 1337, SHA256_K, WState, None, InitialHState)[-1])

    addVars = [f"addLast{i}" for i in range(8)]
    addedWithAtEnd = [BitVec(addVars[i], 32)
                      if addVars[i] not in constraints else constraints[addVars[i]] for i in range(8)]

    all_eqs = []
    for j in range(len(targets)):
        for i in range(8):
            all_eqs.append(((eqs[j][i] + addedWithAtEnd[i]) & 0xffffffff) == targets[j][i])

    _, known_values = solve(all_eqs)

    return known_values


if __name__ == '__main__':

    # before running the script, ssh with provided account
    # go on http://163.172.99.233:8080/956a07cd264c1df26beedcef1b3187ad, start instance
    # ssh mafioso-dev@163.172.99.233 -L 5000:hsmm-bbc3c188:5000 -p 38374 -L 8765:mafiachat-bbc3c188:8000

    steps = [
        (30, 15, [f'addLast{i}' for i in [1, 2, 3, 5, 6, 7]]),
        (28, 14, ['WX59', 'WX60', 'WX61', 'WX62', 'WX63', 'addLast0', 'addLast4']),  # les 3 derniers sont faux mais pg
        (26, 14, ['WX56', 'WX57', 'WX58']),
        (25, 14, ['WX53', 'WX54', 'WX55']),
        (24, 14, ['WX51', 'WX52']),
        (23, 14, ['WX49', 'WX50']),
        (22, 14, ['WX47', 'WX48']),
    ]

    last_steps = [
        (29, 14, ['addLast0', 'addLast4']),
        (0, 14, [f'H{i}' for i in range(8)])
    ]

    constraints = {}

    for step in steps:
        fault_location, n_rounds, to_keep = step

        for_each_fault_location = {
            fault_location: [output_to_abcdefgh(r) for r in do_for(fault_location, 3, n_rounds)]
        }

        print(f"Solving with constraints {constraints}")
        current = solve_corpus(for_each_fault_location, constraints)
        constraints.update({
            k: current[k] for k in to_keep
        })


    WState = [BitVec(f"W{i}", 32) for i in range(16)]
    for _ in range(16, 64):
        n = nextW(WState[-16:])
        WState.append(n)
    all_eqs = []
    for i in range(47, 63):
        all_eqs.append(WState[i] == constraints[f"WX{i}"])

    out, per_key = solve(all_eqs)


    W = [per_key[f"W{i}"] for i in range(16)]
    for i in range(16, 64):
        n = nextW(W[-16:])
        W.append(n)

    for i in range(64):
        constraints[f"WX{i}"] = W[i]
    del constraints['addLast0']
    del constraints['addLast4']

    print(constraints)
    for k, v in constraints.items():
        print(k, hex(v))

    for step in last_steps:
        fault_location, n_rounds, to_keep = step
        for_each_fault_location = {
            fault_location: [output_to_abcdefgh(r) for r in do_for(fault_location, 3, n_rounds)]
        }
        current = solve_corpus(for_each_fault_location, constraints)
        constraints.update({
            k: current[k] for k in to_keep
        })

    print(constraints)
    for k, v in constraints.items():
        print(k, hex(v))

    HFirst = [constraints[f"H{i}"] for i in range(8)]
    print("[+] Solved the desired initial state :)")
    print_hex(HFirst)

    print("[+] We also got hashed message:")
    print_hex(W)

    # hash de notre hash((key||pad) + message)
    to_hash = W
    current_state = HFirst

    out = compute_one_H(0, 32, -1, SHA256_K, to_hash, fault=None, initial_state=current_state)
    test_case = [(a+b) &0xffffffff for a, b in zip(out[0], out[-1])]
    print("[.] Should give the initial legit hash")
    print(format_state_to_hash(test_case))


    rogue = b'\\from:@mafiaDEV\\to:@mafiaBOSS'
    def bytes_to_state(b):
        target = b + b'\x80'
        target = target + b'\x00'*(4-(len(target)%4))
        out = []
        for i in range(0, len(target), 4):
            out.append(int(target[i:i+4].hex(), 16))
        return out
    to_int = bytes_to_state(rogue)
    print_hex(to_int)
    to_int = [*to_int, *[0 for _ in range(15-len(to_int))]]
    to_hash = [
        *to_int, 8 * (0x80 + len(rogue))
    ]
    print("[.] Will hash rogue set as block as follows:")
    print_hex(to_hash)
    for i in range(16, 64):
        n = nextW(to_hash[-16:])
        to_hash.append(n)
    current_start = W[:8]  # the state of next compression (rogue) is also the leak in W of the last compression
    out = compute_one_H(0, 32, -1, SHA256_K, to_hash, fault=None, initial_state=current_start)
    my_hash = [(a + b) & 0xffffffff for a, b in zip(out[0], out[-1])]
    print("[+] We computed the rogue hash:")
    print_hex(my_hash)

    to_hash = [
        *my_hash,
        0x80000000, 0, 0, 0,
        0, 0, 0, 0x00000300
    ]
    for i in range(16, 64):
        n = nextW(to_hash[-16:])
        to_hash.append(n)
    current_state = HFirst

    out = compute_one_H(0, 32, -1, SHA256_K, to_hash, fault=None, initial_state=current_state)
    final_out = [(a + b) & 0xffffffff for a, b in zip(out[0], out[-1])]
    sig = bytes.fromhex(format_state_to_hash(final_out))
    initial = b'from:@mafiaTEST\\to:@nothing\\content:QQo='
    msg = initial + b'\x80' + b'\x00'*(0x40-3-len(initial)) + struct.pack('>H', 8*(0x40+len(initial)))
    msg += rogue
    print(f"Will send rogue message of length {len(msg)}: {msg.hex()}")
    print(base64.b64encode(msg).decode())
    print(base64.b64encode(sig).decode())

    print(f"Will send initial message of length {len(msg[:len(initial)])}: {msg[:len(initial)].hex()}")
    print(base64.b64encode(msg[:len(initial)]).decode())
    print(base64.b64encode(bytes.fromhex(format_state_to_hash(test_case))).decode())


    def test(msg, sig):
        r = remote("127.0.0.1", 5000)
        r.sendline(b"1")
        print(base64.b64encode(sig).decode())
        print(msg)
        r.sendline(base64.b64encode(sig))
        r.sendline(msg)
        data = r.recvuntil(b"\n")
        print(data)
        r.close()

    test(msg[:len(initial)], bytes.fromhex(format_state_to_hash(test_case)))
    test(msg, sig)

    print(base64.b64encode(sig).decode())
    print(base64.b64encode(msg).decode())

    import urllib.parse
    encoded_sig = urllib.parse.quote(base64.b64encode(sig).decode(), safe='')
    encoded_msg = urllib.parse.quote(base64.b64encode(msg).decode(), safe='')
    print(f"curl http://127.0.0.1:8765/msgs -d \"sig={encoded_sig}&data={encoded_msg}\"")
