from utils import nextW, rotate_right

from z3 import BitVec, BitVecVal


WState = [BitVec(f"WX{i}", 32) for i in range(16)]
InitialHState = [BitVec(f"H{i}", 32) for i in range(8)]
addedWithAtEnd = [BitVec(f"addLast{i}", 32) for i in range(8)]
faults = [BitVec(f"F{i}", 32) for i in range(32)]
toAndWith = BitVecVal(0xffffffff, 32)

all_vars = [*WState, *InitialHState, *addedWithAtEnd, *faults]



def compute_all_W():
    for _ in range(16, 64):
        n = nextW(WState[-16:])
        WState.append(n)

compute_all_W()



def nextH_double(state, k1, w1, k2, w2, is_fault=None):
    a, b, c, d, e, f, g, h = state
    c = c if is_fault is None else (c + is_fault)

    s1 = rotate_right(e, 6) ^ rotate_right(e, 11) ^ rotate_right(e, 25)
    ch = (e & f) ^ ((~e) & g)
    t1 = (h + s1 + ch + k1 + w1) & 0xffffffff
    s0 = rotate_right(a, 2) ^ rotate_right(a, 13) ^ rotate_right(a, 22)
    maj = ((a & b) ^ (a & c) ^ (b & c)) & 0xffffffff
    t2 = (s0 + maj) & 0xffffffff

    a2, b2, c2, d2, e2, f2, g2, h2 = [(t1 + t2) & 0xffffffff, a, b, c & 0xffffffff, (d + t1) & 0xffffffff, e, f, g]


    s1 = rotate_right(e2, 6) ^ rotate_right(e2, 11) ^ rotate_right(e2, 25)
    ch = (e2 & f2) ^ (~e2 & g2)
    tmp1 = (g + s1 + ch + k2 + w2) & 0xffffffff
    s0 = rotate_right(a2, 2) ^ rotate_right(a2, 13) ^ rotate_right(a2, 22)
    maj = (a2 & a) ^ (a2 & b) ^ (a & b)  # there is the trick
    tmp2 = (tmp1 + s0 + maj) & 0xffffffff
    tmp3 = (d2 + tmp1) & 0xffffffff

    return (tmp2, a2, tmp3, e2)




def compute_all_H(fault_location, K, WState, n_steps=16):
    currentHState = InitialHState
    all_states = [InitialHState]
    for i in range(n_steps):
        w1, w2, w3, w4 = WState[4*i: 4*(i+1)]
        k1, k2, k3, k4 = K[4*i: 4*(i+1)]

        a, b, c, d, e, f, g, h = currentHState
        na, nb, ne, nf = nextH_double(currentHState, k1, w1, k2, w2,
                                      faults[fault_location] if fault_location == 2*i else None)
        currentHState = (na, nb, a, b, ne, nf, e, f)
        all_states.append(currentHState)

        a, b, c, d, e, f, g, h = currentHState
        na, nb, ne, nf = nextH_double(currentHState, k3, w3, k4, w4,
                                      faults[fault_location] if fault_location == 2*i + 1 else None)
        currentHState = (na, nb, a, b, ne, nf, e, f)
        all_states.append(currentHState)

    return all_states


def compute_one_H(begin, end, fault_location, K, WState, fault, initial_state=None):
    currentHState = InitialHState if not initial_state else initial_state
    all_states = [currentHState]
    for i in range(begin, end):
        w1, w2 = WState[2 * i: 2 * (i + 1)]
        k1, k2 = K[2 * i: 2 * (i + 1)]

        a, b, c, d, e, f, g, h = currentHState
        na, nb, ne, nf = nextH_double(currentHState, k1, w1, k2, w2,
                                            fault if fault_location == i else None)
        currentHState = (na, nb, a, b, ne, nf, e, f)
        all_states.append(currentHState)
    return all_states
