from action.utils.solve_constraints import solve_z3_problem

from z3 import LShR, And
import hashlib


HOST = '127.0.0.1'
PORT = 5000


SHA256_K = [
    0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1,
    0x923f82a4, 0xab1c5ed5, 0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3,
    0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174, 0xe49b69c1, 0xefbe4786,
    0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
    0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147,
    0x06ca6351, 0x14292967, 0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13,
    0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85, 0xa2bfe8a1, 0xa81a664b,
    0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
    0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a,
    0x5b9cca4f, 0x682e6ff3, 0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,
    0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
]

def rotate_right(v, n):
    w = LShR(v, n) | (v << (32 - n)) if not isinstance(v, int) else (v >> n) | (v << (32 - n))
    return w & 0xffffffff

def sum0(W):
    return rotate_right(W, 7) ^ rotate_right(W, 18) ^ (LShR(W, 3) if not isinstance(W, int) else (W >> 3))

def sum1(W):
    return rotate_right(W, 17) ^ rotate_right(W, 19) ^ (LShR(W, 10) if not isinstance(W, int) else (W >> 10))


def nextW(sixteenLatestW):
    return (sixteenLatestW[-16] + sum0(sixteenLatestW[-15]) + sixteenLatestW[-7] + sum1(sixteenLatestW[-2])) & 0xffffffff


def ordered(x):
    f, e, b, a, h, g, d, c = x
    return [a, b, c, d, e, f, g, h]

def output_to_abcdefgh(output):
    target = []
    for j in range(0, len(output), 8):
        target.append(int(output[j: j+8], 16))
    return target

def print_hex(data):
    print(list(map(hex, data)))


def try_parse(data):
    def interesting_lines():
        lines = data.split("\n")
        for l in lines:
            if '|H' in l or '|W' in l or '|addLast' in l or '|F' in l:
                yield l
    interesting = list(interesting_lines())
    def cut(l):
        key = l.split('|')[1]
        value = l.split('#x')[-1].split(' ')[0]
        return (key, int('0x'+value, 16))
    return sorted(map(cut, interesting))


def solve(eqs):
    content = solve_z3_problem(And(*eqs))['stdout']
    print(content)
    out = try_parse(content)
    per_key = {}
    for k, v in out:
        print(f"[+] RECEIVED FROM SOLVER: {k}, {hex(v)}")
        per_key[k] = v
    return out, per_key


def MD5(data):
    h = hashlib.md5()
    h.update(data)
    return h.hexdigest()


def format_state_to_hash(T):
    return ''.join(map(lambda x: hex(x)[2:].zfill(8), T))
