
const remote = "MAIN_URL/";


class DoAction {

  constructor(defaultDelay, readMem, writeMem, flushWrites, crash, trigger, setHighBits, loopRaceCondition) {
    this.defaultDelay = defaultDelay;
    this.readMem = readMem;
    this.writeMem = writeMem;
    this.crash = crash;
    this.raceCondition = loopRaceCondition;
    this.trigger = trigger;
    this.flushWrites = flushWrites;
    this.setHighBits = setHighBits;
    this.state = {};   // all known values + size read
    this.orders = {};  // all received orders
    this.reinitResults();
    this.hasTriggered = false;
  }

  reinitResults () {
    this.results = {'read': [], 'write': [], 'delay': []};
  }

  makeRequest (action, done, params) {
    var xhr = new XMLHttpRequest();
    xhr.open('POST', remote + ((action === 'ask' || action === 'error') ? 'comm/' : 'do/') + action);
    xhr.onload = function () {
      done && done(null, JSON.parse(xhr.response));
    };
    xhr.onerror = function () {
      done && done(xhr.response);
    };
    xhr.setRequestHeader("Content-Type", "application/json;charset=UTF-8");
    xhr.send(JSON.stringify(params));
  }

  sendError (error) {
    error && this.makeRequest('error', undefined, {error: error});
  }


  mainLoop () {
    this.i1 = setInterval(() => this.askAndQueueNewActions(), this.defaultDelay);
    this.i2 = setInterval(() => this.sendObtainedResults(), this.defaultDelay);
  }


  askAndQueueNewActions () {
    var self = this;
    this.makeRequest('ask', function (err, data) {
      console.log("Data received");
      console.log(data)
      err && self.sendError(err);
      data && data.map((order) => {
          self.dispatch(order.type, order)
        })
    })
  }

  sendObtainedResults () {
    var previous = this.results;
    var self = this;
    this.reinitResults();
    Object.keys(previous).map((key) => {
      previous[key]?.length > 0 &&
      this.makeRequest(key, function (err, data) {
        err && self.sendError(err);
      }, previous[key])
    })
  }


  dispatch(type, input) {
    console.log(input)
    switch (type) {
      case 'read':
        return this.readAt({...input});
        break;
      case 'write':
        return this.writeAt({...input});
        break;
      case 'delay':
        return this.setDelay({...input});
        break;
      case 'stop':
        return this.stop({...input});
        break;
      case 'crash':
        return this.crash({...input});
        break;
      case 'trigger':
        if (!this.hasTriggered) {
            this.hasTriggered = true;
            setTimeout(() => {
              console.log("Flushing writes");
              this.flushWrites();
            }, 2000);
            setTimeout(() => {
              console.log("Triggering");
              this.trigger({...input});
            }, 5000);
            this.setDelay({delay: 10000});
        }
        else {
            setTimeout(() => {
              console.log("pull the trigger back");
              this.hasTriggered = false;
            }, 10000);
        }
        break;
      case 'highbits':
        return this.setHighBits({...input});
        break;
      /*case 'race':
        i = this.raceCondition({...input});
        console.log("two tasks");
        console.log(i);
        break;*/
      case 'fakebuf':
        return this.setFakeBuf({...input});
        break;
    }
  }

  readAt({addr, size}) {
    console.log(addr)
    console.log(this);
    var result = this.readMem({addr, size});
    console.log("READ AT "+addr+ " "+size);
    console.log({addr: addr, size: result.length, result: 'SUCCESS', hexcontent: buf2hex(result)})
    this.results['read'].push({addr: addr, size: result.length, result: 'SUCCESS', hexcontent: buf2hex(result)});
    console.log(this.results)
  }

  writeAt({addr, content}) {
    console.log("WRITE AT "+addr+" "+content);
    this.writeMem({addr, content});
  }

  setDelay({delay}) {
    console.log("SETTING DELAY "+delay)
    clearInterval(this.i1);
    clearInterval(this.i2);
    clearInterval(this.i3);
    this.defaultDelay = delay;
    this.mainLoop();
    this.results['delay'].push({'success': 'SUCCESS'})
  }

  stop() {
    clearInterval(this.i1);
    clearInterval(this.i2);
    clearInterval(this.i3);
    console.log("ENDING");
  }
}
