from pydantic import BaseModel

from typing import Literal
from enum import Enum


class ReadOrder(BaseModel):
    type: Literal['read'] = 'read'
    addr: str
    size: int = 8
    is_absolute: bool = True

class WriteOrder(BaseModel):
    type: Literal['write'] = 'write'
    addr: str
    content: str
    relative: bool = True

class StopOrder(BaseModel):
    type: Literal['stop'] = 'stop'

class CrashOrder(BaseModel):
    type: Literal['crash'] = 'crash'

class SwitchDelay(BaseModel):
    type: Literal['delay'] = 'delay'
    newDelay: float

class FinalTrigger(BaseModel):
    type: Literal['trigger'] = 'trigger'

class ReceiveHigh(BaseModel):
    type: Literal['highbits'] = 'highbits'
    addr: str

class EnterRaceCondition(BaseModel):
    type: Literal['race'] = 'race'
    addr: str

Order = ReadOrder | WriteOrder | StopOrder | SwitchDelay | CrashOrder | EnterRaceCondition | FinalTrigger | ReceiveHigh



class ResultEnum(str, Enum):
    success = "SUCCESS"
    failure = "FAILURE"


class ReadUnitIn(BaseModel):
    addr: str
    size: int = 0
    hexcontent: str
    result: ResultEnum

class WriteUnitIn(BaseModel):
    addr: str
    size: int
    result: ResultEnum

class SetDelayIn(BaseModel):
    result: ResultEnum
