from core.core00_core_model.mixin.base_mixin import BaseMixinsNoJoin
from core.core00_core_model.concept.timed import CreatedAt

from .with_context import WithContext

from sqlalchemy import Integer, BIGINT, Boolean, UniqueConstraint
from sqlalchemy.orm import relationship, mapped_column, Mapped


class ReadQuery(*BaseMixinsNoJoin, WithContext, CreatedAt):

    __tablename__ = 'ReadQuery'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    # address = mapped_column(BIGINT)  # cannot do on SQLite due to 2**63-1 ...
    low32: Mapped[int] = mapped_column(BIGINT)
    high32: Mapped[int] = mapped_column(BIGINT)
    size: Mapped[int] = mapped_column(Integer)
    is_absolute: Mapped[bool] = mapped_column(Boolean, default=True)

    answer = relationship('ReadQueryAnswer', back_populates='read_query')

    __table_args__ = (UniqueConstraint('low32', 'high32', 'size', 'local_context_id', name='unique_read_query_local'),
                      UniqueConstraint('low32', 'high32', 'size', 'remote_context_id',
                                       name='unique_read_query_remote'),)

    def __init__(self, address=0, is_absolute=False, size=0, local_context=None, remote_context=None):
        super().__init__(low32=address & 0xffffffff,
                         high32=(address >> 32) & 0xffffffff,
                         is_absolute=is_absolute,
                         size=size,
                         local_context=local_context,
                         remote_context=remote_context)

    @property
    def address(self):
        return self.low32 + (self.high32 << 32)

    @address.setter
    def address(self, address):
        self.low32 = address & 0xffffffff
        self.high32 = (address >> 32) & 0xffffffff

    @property
    def type_read(self):
        return 'ABS:' if self.is_absolute else 'REL:'

    @classmethod
    def search_query(cls, address, size, context):
        return cls.filter_by_context(context).filter_by(
            low32=address & 0xffffffff,
            high32=(address >> 32) & 0xffffffff,
            size=size
        ).one_or_none()

    def __repr__(self):
        return f"Query for {hex(self.address)} size {hex(self.size)} (in context " \
               f"{self.local_context if self.local_context else self.remote_context})"
