from sqlalchemy.orm import Mapped, mapped_column, relationship, declared_attr
from sqlalchemy import Integer, ForeignKey

from wussticcommon.model.local_sstic_context import LocalSSTICContext
from wussticcommon.model.remote_sstic_context import RemoteSSTICContext


class WithContext:
    __abstract__ = True

    local_context_id: Mapped[int] = mapped_column(Integer, ForeignKey(LocalSSTICContext.id), nullable=True)
    remote_context_id: Mapped[int] = mapped_column(Integer, ForeignKey(RemoteSSTICContext.id), nullable=True)

    @declared_attr
    def local_context(cls) -> Mapped[LocalSSTICContext]:
        return relationship(LocalSSTICContext, foreign_keys=[cls.local_context_id])

    @declared_attr
    def remote_context(cls) -> Mapped[RemoteSSTICContext]:
        return relationship(RemoteSSTICContext, foreign_keys=[cls.remote_context_id])

    @classmethod
    def filter_by_context(cls, context: LocalSSTICContext | RemoteSSTICContext):
        match context:
            case LocalSSTICContext():
                return cls.query.filter_by(local_context=context)
            case RemoteSSTICContext():
                return cls.query.filter_by(remote_context=context)
            case _:
                raise NotImplementedError

    @classmethod
    def context_filter(cls, context: LocalSSTICContext | RemoteSSTICContext):
        match context:
            case LocalSSTICContext():
                return cls.local_context == context
            case RemoteSSTICContext():
                return cls.remote_context == context
            case _:
                raise NotImplementedError

    def search_query(self):
        return self.local_context and {'local_context': self.local_context} or {'remote_context': self.remote_context}


def search_query(ctxt):
    match ctxt:
        case LocalSSTICContext():
            return {'local_context': ctxt}
        case RemoteSSTICContext():
            return {'remote_context': ctxt}
        case _:
            raise NotImplementedError
