from core.core00_core_model.mixin.base_mixin import BaseMixinsNoJoin
from core.core00_core_model.concept.timed import CreatedAt

from .with_context import WithContext

from sqlalchemy.orm import relationship, Mapped, mapped_column
from sqlalchemy import Integer, String, UniqueConstraint


class WriteQuery(*BaseMixinsNoJoin, WithContext, CreatedAt):

    __tablename__ = 'WriteQuery'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    #address = Column(BIGINT)  # cannot do on SQLite due to 2**63-1
    address: Mapped[String] = mapped_column(String)
    content: Mapped[String] = mapped_column(String)

    answer = relationship('WriteQueryAnswer', back_populates='write_query')

    __table_args__ = (UniqueConstraint('address', 'content', 'local_context_id', name='unique_write_query_local'),
                      UniqueConstraint('address', 'content', 'remote_context_id', name='unique_write_query_remote'),)

    @classmethod
    def search_query(cls, address, content, context):
        return cls.filter_by_context(context).filter_by(address=address, content=content).one_or_none()

    def __repr__(self):
        return f"Query for writing at {self.address} content {self.content} (in context" \
               f" {self.local_context if self.local_context else self.remote_context})"
