from core.core05_persistent_model.policy.session import get_session

from ..api.models import ReadUnitIn, ReadOrder, StopOrder, WriteOrder, CrashOrder, EnterRaceCondition, ReceiveHigh, \
    FinalTrigger
from .pwn_context import PwnContext
from .pwn_utils import get_confirmed_context

from ..model.read_query_answer import ReadQueryAnswer
from ..model.read_query import ReadQuery

from typing import List


def get_next_actions():
    with get_session():
        ctxt = get_confirmed_context()
        pwn_ctxt = PwnContext(ctxt, True)  # we ping current server

        pwn_ctxt.minimal_check()
        crash_query = pwn_ctxt.next_crash_query()
        stop_query = pwn_ctxt.next_stop_query()
        read_queries = pwn_ctxt.next_read_queries()
        write_queries = pwn_ctxt.next_write_queries()
        set_high_bits_query = pwn_ctxt.next_set_high_bit()
        trigger_query = pwn_ctxt.next_trigger_query()

        read_orders = [ReadOrder(addr=hex(rq.address), size=rq.size) for rq in read_queries]
        write_orders = [WriteOrder(addr=wq.address, content=wq.content) for wq in write_queries]
        stop_orders = [StopOrder()] if stop_query else []
        crash_orsers = [CrashOrder()] if crash_query else []
        highbits_orders = [ReceiveHigh(addr=hex(set_high_bits_query))] if set_high_bits_query else []
        trigger_orders = [FinalTrigger()] if trigger_query else []
        return [*read_orders, *write_orders, *stop_orders, *crash_orsers, *highbits_orders, *trigger_orders]


def receive_leak(data: List[ReadUnitIn] | ReadUnitIn):
    with get_session():
        ctxt = get_confirmed_context()
        match data:
            case ReadUnitIn():
                print(f"[.] Received read: {data}")
                total_size = len(data.hexcontent)//2

                base_query = ReadQuery.search_query(int(data.addr, 16), total_size, ctxt)
                if not base_query:
                    print(f"[-] Don't know why but the query record for {data.addr} (size {hex(data.size)}) "
                          f"does not exist")
                    return
                else:
                    existing = ReadQueryAnswer.get_for(read_query=base_query)
                    if not existing:
                        query_answer = ReadQueryAnswer.create(read_query=base_query,
                                                              content=bytes.fromhex(data.hexcontent))
                    else:
                        if existing.content == bytes.fromhex(data.hexcontent):
                            print(f"[+] Already existing answer for {base_query}, content matching")
                        else:
                            print(f"[-] Already existing answer for {base_query}, content not matching please investigate")
            case list():
                for subdata in data:
                    receive_leak(data=subdata)
            case _:
                raise NotImplementedError

    return {}
