from core.core30_context.context_dependency_graph import context_producer
from core.core11_config.config import config_dependencies, Config
from core.core30_context.context import Context

from .manage_local import execute_winrm, execute_winrm_interactive, working_local_context
from .manage_remote import remote_command, remote_interactive, working_remote_context

from typing import List


class Shell:

    def __init__(self, execute_commands, interactive, context):
        self.execute = execute_commands
        self.interactive = interactive
        self.context = context

    def execute(self, commands: str | List[str], *args):
        return self.execute(commands, *args)

    def interactive(self):
        return self.interactive()


#@context_dependencies(('.systemctxt.local_system', ExecContext), ('.systemctxt.remote_system', ExecContext))
@context_producer(('.systemctxt.execute_shell', Shell), ('.systemctxt.execution_context_id', str))
@config_dependencies(('.systemctxt.local_or_remote', bool))
def pwn_shell(config: Config, ctxt: Context):
    if config['systemctxt']['local_or_remote']:
        local_context = working_local_context()
        print(f"[.] Checked local context {local_context.instance_id}")
        # don't take error answer as not supported in remote context (so lambda)
        shell = Shell(lambda *args: execute_winrm(*args)[1],
                      execute_winrm_interactive, {'type': 'LOCAL', 'ctxt': local_context})
        ctxt.setdefault('systemctxt', {})['execute_shell'] = shell
        ctxt['systemctxt']['execution_context_id'] = local_context.instance_id
    else:
        remote_context = working_remote_context()
        print(f"[.] Checked remote context {remote_context.instance_id}")
        shell = Shell(remote_command, remote_interactive, {'type': 'REMOTE', 'ctxt': remote_context})
        ctxt.setdefault('systemctxt', {})['execute_shell'] = shell
        ctxt['systemctxt']['execution_context_id'] = remote_context.instance_id
    return ctxt['systemctxt']['execute_shell']
