from core.core11_config.config import config_dependencies, Config

from wussticcommon.ensure_service.local_instance import get_working_context

from threading import Thread
from typing import List
import subprocess
import select
import psutil
import time
import sys


def execute_process_send_stdin_get_stdout(command, arguments, timeout=60):
    p = subprocess.Popen(command.split(' '), stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    return p.communicate(input=arguments.encode(), timeout=timeout)


@config_dependencies(('.sstic.winrm_command', str))
def winrm_command(config: Config):
    return config['sstic']['winrm_command']

base_command = winrm_command()

script_command = ['script', '-q', '/dev/null', *base_command.split()]

def execute_winrm_interactive():
    p = subprocess.Popen(script_command,
                         stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout = p.stdout.fileno()
    stderr = p.stderr.fileno()
    stdout_data = b''
    new_out_data = b''
    last_data = time.time()
    while True:
        sockets_list = [stdout]
        read_sockets, write_sockets, error_sockets = select.select(sockets_list, [], [], 0.1)

        for fd in read_sockets:
            if fd == sys.stdin:
                read = sys.stdin.readline()
                p.stdin.write(read.encode())
            elif fd == stdout or fd == stderr:
                f = p.stdout if fd == stdout else p.stderr
                new_out_data += f.read(1)
                last_data = time.time()

        if (time.time() - last_data > 0.1) and new_out_data:
            prev, new_out_data = new_out_data, b''
            stdout_data += prev
            sys.stdout.write(prev.decode('latin1'))

        if not new_out_data:
            time.sleep(0.03)

        if p.poll() is not None:
            break

    return stdout_data


def execute_winrm(command: str | List, timeout=120):
    match command:
        case str():
            arguments = command + '\nexit\n'
        case _:
            arguments = '\n'.join(command) + '\nexit\n'
    stdout, stderr = execute_process_send_stdin_get_stdout(base_command, arguments, timeout)
    is_error = b'Exiting with code 1' in stdout
    return is_error, stdout.split(b'connection to remote endpoint')[-1].replace(b'\r\r\n', b'\n')



def check_local_running():
    # check pid of qemu-system
    if len([p.name() for p in psutil.process_iter() if 'qemu' in p.name()]) == 0:
        return False
    # check answer of execute winrm locally
    _, out = execute_winrm('whoami')
    if b'enigma' in out:
        return True
    return False


def check_service_running():
    print("[.] Checking service running (local winrm)")
    error, out = execute_winrm('Get-Service -Name netshdw')
    if not error and b'Running' in out:
        return error, True
    return error, False


def start_service():
    print("[.] Start service (local winrm)")
    error, out = execute_winrm('net start netshdw')
    if not error and (b'The netshdw service was started successfully' in out or b'has already been started' in out):
        return error, True
    return error, False


def ensure_ok_or_fail(withdebug):
    _, is_running = check_service_running()
    if is_running:
        print("[+] Service apparently running, continue")
        return True
    else:
        print("[.] Service apparently not running, will start it")
    if not check_local_running():
        t = Thread(target=launch_qemu, args=(withdebug,))
        t.start()
        print("[.] Qemu launched, waiting a bit before checking. Please note the current process will never terminate"
              " until session is complete")
        time.sleep(60)
    error, is_running = check_service_running()
    if error:
        print("[-] Bad news error triggered when checking if all running as expected")
        raise Exception('Failed to start qemu')
    else:
        print("[.] Local service running")
    if not is_running:
        error, _ = start_service()
        if error:
            print("[-] Bad news error triggered when starting the service :/")
            raise Exception('Failed to start service')
        else:
            print("[.] Local service started")
    return True


def working_local_context():
    ctxt = get_working_context()
    ensure_ok_or_fail(False)
    return ctxt


if __name__ == '__main__':
    print(ensure_ok_or_fail(True))

    is_error, out = execute_winrm('whoami')
    print("_____ is_error")
    print(is_error)
    print("_____ out")
    print(out)

    err, status = check_service_running()
    print(f"Is running ? {status}")
    if not status:
        print(start_service())
