from core.core30_context.context_dependency_graph import context_dependencies, context_producer
from core.core30_context.context import Context

from wussticcommon.ensure_service.instance_lifecycle import new_context
from wussticcommon.model.remote_sstic_context import RemoteSSTICContext

from socket import socket, AF_INET, SOCK_STREAM
from typing import List
import select
import time
import sys


def init_conn(ip, port):
    s = socket(AF_INET, SOCK_STREAM)
    s.settimeout(5)
    s.connect((ip, port))
    s.send(b"FYTn5OybJ097HUOr0oQSIja6pN4aagTj\n")
    return s


@context_dependencies(('.targetctxt.valid_remote_context', RemoteSSTICContext))
def remote_command(ctxt: Context, command: str | List, timeout=20, print_stdout=True):
    s = init_conn(ctxt['targetctxt']['valid_remote_context'].ip, ctxt['targetctxt']['valid_remote_context'].port2)
    match command:
        case str():
            s.send((command + '\n').encode())
            time.sleep(0.1)
        case _:
            for c in command:
                s.send((c + '\n').encode())
                time.sleep(0.1)

    answer = b''
    start = time.time()
    finished = False
    while not finished:
        socket_list = [s]
        read_sockets, write_sockets, error_sockets = select.select(socket_list, [], [], 1)
        for sock in read_sockets:
            if sock == s:
                try:
                    data = sock.recv(1024)
                except:
                    finished = True
                if not data:
                    finished = True
                    break
                else:
                    if print_stdout:
                        sys.stdout.write(data.decode())
                    answer += data
        if timeout and time.time() - start > timeout:
            print("[.] Time-outed")
            break
        time.sleep(0.1)
        if not finished:
            try:
                s.send('sleep 1\nexit\n'.encode())
            except:
                pass
    s.close()
    return answer.replace(b'\r\r\n', b'\n')


@context_dependencies(('.targetctxt.valid_remote_context', RemoteSSTICContext))
def remote_interactive(ctxt: Context):
    s = init_conn(ctxt['targetctxt']['valid_remote_context'].ip, ctxt['targetctxt']['valid_remote_context'].port2)
    finish = False
    while not finish:
        socket_list = [sys.stdin, s]
        read_sockets, write_sockets, error_sockets = select.select(socket_list, [], [], 1)

        for sock in read_sockets:
            if sock == s:
                data = sock.recv(1024)
                if not data:
                    finish = True
                    break
                else:
                    sys.stdout.write(data.decode('latin1'))
            else:
                msg = sys.stdin.readline()
                s.send(msg.encode())
    s.close()


def check_remote_running():
    out = remote_command('whoami')
    return 'enigma' in out.decode('latin1')


def check_service_running():
    out = remote_command('Get-Service -Name netshdw')
    return 'Running' in out.decode('latin1')


@context_dependencies(('.targetctxt.valid_remote_context', RemoteSSTICContext),  ('.targetctxt.step_secret_url', str))
@context_producer(('.systemctxt.remote_system', RemoteSSTICContext))
def working_remote_context(ctxt: Context):
    context = ctxt['targetctxt']['valid_remote_context']
    ctxt.setdefault('systemctxt', {})['remote_system'] = context
    ip = ctxt['targetctxt']['valid_remote_context'].ip
    port = ctxt['targetctxt']['valid_remote_context'].port2
    print(f"[.] Checking TCP context at {ip}:{port}")
    if not check_remote_running():
        try:
            context = new_context(context.step, ctxt['targetctxt']['step_secret_url'])
            ctxt['targetctxt']['valid_remote_context'] = context
            ctxt['systemctxt']['remote_system'] = context
            if not check_remote_running():
                raise Exception(f"Bad context {context}, whoami failed")
        except Exception as e:
            raise e
    else:
        print("[+] Whoami OK")
    if not check_service_running():
        raise Exception(f"Bad context {context}, get-service failed")
    else:
        print("[+] Service OK")
    return ctxt['systemctxt']['remote_system']


if __name__ == '__main__':
    from core.core05_persistent_model.policy.session import get_session

    with get_session():
        print(remote_command("whoami /priv"))
        remote_interactive()
