from core.core30_context.context_dependency_graph import context_dependencies
from core.core11_config.config import config_dependencies, Config
from core.core30_context.context import Context
from core.core02_model.typed.file import File

from action.utils.control_server import upload_on_controlled

from manage import Shell


# For upload:
# we check we have ssh access to the provided server in config
# if we do not have, we generate a key and wait for it to be pushed
# when ok, scp-ing the desired file with paramiko
# then launching command on powershell to download from url


@context_dependencies(('.systemctxt.execute_shell', Shell),)
@config_dependencies(('.systemctxt.controlled.http_service', str),
                     ('.systemctxt.remote_upload_dir', str))
def upload_data(config: Config, ctxt: Context, content: File, filename: str | None, write_copy: bool = True):
    _ = upload_on_controlled(content, filename, write_copy)
    # then trigger the pull order
    url = f"{config['systemctxt']['controlled']['http_service']}/{filename}"
    remote_fname = config['systemctxt']['remote_upload_dir'] + '\\' + filename
    command = f'Invoke-WebRequest -Uri "{url}" -Out {remote_fname}'
    print(f"[.] Will execute {command} on the remote location")
    out = ctxt['systemctxt']['execute_shell'].execute(command)

    print(f"[+] Upload finished, check {remote_fname}")
    return out


@context_dependencies(('.systemctxt.execute_shell', Shell))
def ensure_file(ctxt: Context, fullname):
    fname_split = fullname.split('\\')
    if len(fname_split) <= 1:
        dirname = '.'
        fname = fname_split[-1]
    else:
        dirname = '\\'.join(fname_split[:-1])
        fname = fname_split[-1]
    dir_command = f"dir {fullname}"
    out = ctxt['systemctxt']['execute_shell'].execute(dir_command)

    expected1 = f"Directory: {dirname if dirname != '.' else ''}"
    expected2 = fname
    if expected1.encode() not in out or expected2.encode() not in out:
        print(f"[.] File {fullname} not found")
        return False
    return True


if __name__ == '__main__':
    from core.core05_persistent_model.policy.session import get_session
    from core.core02_model.typed.file import FileContent

    with get_session():
        print(upload_data(content=FileContent(content=b'test'), filename='out.txt'))
