#include "getline.h"

#include <ctype.h>
#include <errno.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include "error.h"

#define MAX_INPUT_SZ UINT16_MAX

size_t hsm_getline(char **lineptr) {
  if (*lineptr != NULL) {
    hsm_error(ERROR_INTERNAL);
  }

  size_t n = 0;
  ssize_t ssz = getline(lineptr, &n, stdin);

  if (ssz < 0 || MAX_INPUT_SZ < ssz) {
    hsm_error(ERROR_INPUT);
  }
  size_t sz = (size_t)ssz;

  if (1 <= sz && (*lineptr)[sz - 1] == '\n') {
    (*lineptr)[sz - 1] = '\0';
    sz--;
  }

  return sz;
}

void hsm_getline_check(char *in, size_t sz, bool must_start_with_at,
                       bool base64_ok) {
  size_t i = 0;

  if (must_start_with_at) {
    if (sz < 1 || in[0] != '@') {
      hsm_error(ERROR_INPUT);
    }
    i = 1;
  }

  while (i < sz) {
    char c = in[i];
    i++;

    if (isalnum(c) != 0) {
      continue;
    }

    if (base64_ok && (c == '+' || c == '/' || c == '=')) {
      continue;
    }

    hsm_error(ERROR_INPUT);
  }
}

unsigned long hsm_getline_nb() {
  char *nb_str = NULL;
  size_t sz = hsm_getline(&nb_str);
  hsm_getline_check(nb_str, sz, false, false);
  errno = 0;
  char *endptr = NULL;
  long nb_l = strtol(nb_str, &endptr, 10);
  if (errno != 0 || *endptr != '\0' || nb_l < 0) {
    hsm_error(ERROR_INPUT);
  }
  free(nb_str);
  return (unsigned long)nb_l;
}
