#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include "auth.h"
#include "csr.h"
#include "error.h"
#include "getline.h"
#include "message.h"

void main_sign() {
  char *username = NULL;
  size_t username_sz = hsm_getline(&username);
  hsm_getline_check(username, username_sz, true, false);

  char *password = NULL;
  size_t password_sz = hsm_getline(&password);
  hsm_getline_check(password, password_sz, false, false);

  char *recipient = NULL;
  size_t recipient_sz = hsm_getline(&recipient);
  hsm_getline_check(recipient, recipient_sz, true, false);

  char *content = NULL;
  size_t content_sz = hsm_getline(&content);
  hsm_getline_check(content, content_sz, false, true);

  unsigned long extra_0 = hsm_getline_nb();
  csrw_mafia_0(extra_0);

  unsigned long extra_1 = hsm_getline_nb();
  csrw_mafia_1(extra_1);

  auth(username, password);

  char *message = NULL;
  size_t message_sz = format(&message, username, username_sz, recipient,
                             recipient_sz, content, content_sz);
  char *message_b64 = NULL;
  encode(&message_b64, (void *)message, message_sz);

  char signature_b64[SIG_SZ_B64 + 1]; // + 1 for /0
  sign(signature_b64, (void *)message, message_sz);

  printf("%s\n%s\n", message_b64, signature_b64);
  fflush(stdout);

  free(username);
  free(password);
  free(recipient);
  free(content);
  free(message);
  free(message_b64);

  exit(EXIT_SUCCESS);
}

void main_verify() {
  char *claim = NULL;
  size_t claim_sz = hsm_getline(&claim);
  hsm_getline_check(claim, claim_sz, false, true);
  if (claim_sz != SIG_SZ_B64) {
    hsm_error(ERROR_INPUT);
  }

  char *message = NULL;
  size_t message_sz = hsm_getline(&message);
  // No need to call hsm_getline_check since we only check signature,
  // also message should contain caracters like ':' and '\'

  printf("%i\n", verify(claim, (void *)message, message_sz));
  fflush(stdout);

  free(claim);
  free(message);

  exit(EXIT_SUCCESS);
}

int main() {
  csrw_mafia_0(1337);
  csrw_mafia_1(1337);
  csrw_mafia_1_now(0);
  unsigned long choice = hsm_getline_nb();
  if (choice == 0) {
    main_sign();
  } else if (choice == 1) {
    main_verify();
  }
  hsm_error(ERROR_INPUT);
}
