from core.core30_context.context_dependency_graph import context_dependencies
from core.core30_context.context import Context

from step1_utils_interact import SERVER_HOST, SERVER_PORT
from step1_model import decryption_success

from typing import Dict
import socket


def send_receive(send):
    r = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    r.settimeout(10)
    r.connect((SERVER_HOST, SERVER_PORT))
    r.send(b"\x01"+send)
    tmp = r.recv(1024)
    r.close()
    return tmp

def decrypt(block):
    cur = b'\x00'*16
    for i in range(16):
        for j in range(256):
            tosend = b'\x00'*(15-i) + bytes([j]) + cur[16-i:]
            out = send_receive(tosend + block)
            if b'ISO 7816-4' not in out and b'Incomplete message' not in out:
                cur = cur[:15-i] + bytes([j ^ 0x80]) + cur[16-i:]
                break
    return cur


def block_of(data, block):
    return data[16*block: 16+16*block]

def xor(d1, d2):
    assert len(d1) == len(d2), f"WARNING LENGTH DIFFER of {d1.hex()} and {d2.hex()}"
    return bytes([x ^ y for x, y in zip(d1, d2)])

def decrypt_one(enc_0, enc_1):
    return xor(enc_0, decrypt(enc_1))


def decrypt_full(data):
    n_block = len(data)//16
    out_for = b''
    for i in range(1, n_block):
        print(f"==== Decryption of block {i} ====")
        b1 = block_of(data, i-1)
        b2 = block_of(data, i)
        out_for += xor(b1, decrypt(b2))
    return out_for


def construct_data_to_send(data):
    padded = data + b"\x80" + b"\x00"*(16-((len(data)+1)%16))
    n_blocks = len(padded)//16
    cur_block = b'\x00'*16
    to_send = b''
    for i in range(len(padded)//16):
        to_send = cur_block + to_send
        decrypted = decrypt(cur_block)
        cur_block = xor(decrypted, block_of(padded, n_blocks-i-1))
    to_send = cur_block + to_send
    return to_send


@context_dependencies(('.exploit.step1.decrypted_matches', Dict[str, str]))
def decrypt_and_act(ctxt: Context, msg_index, data, block_to_take):
    prev = data[block_to_take * 0x10: (block_to_take + 1) * 0x10]
    block = data[(block_to_take + 1) * 0x10: (block_to_take + 2) * 0x10]
    all_decrypted_blocks = ctxt['exploit']['step1']['decrypted_matches']

    if block.hex() in all_decrypted_blocks:
        out = bytes.fromhex(all_decrypted_blocks[block.hex()])
    else:
        out = decrypt_one(prev, block)
        decryption_success(msg_index, block_to_take, out)
    print(f"[.] Decrypted {block.hex()} => {out.hex()}")
    return out


# the idea of this class is to get messages to decrypt in some kind of "optimized" order, doing unknown first
# it yields blocks to decrypt (under the form of the previous enc block and the one to decrypt for xor)
class OrderedMessages:

    def __init__(self, all_known_messages, all_unknwown_messages):
        self.all_known_messages = all_known_messages
        self.all_unknwown_messages = all_unknwown_messages
        self.already_tested = {(msg.message.length, msg.answer and msg.answer.length or 0)
                               for msg in self.all_known_messages}
        self.known_sizes = {msg.message.length for msg in self.all_known_messages}
        self.known_answer_sizes = {msg.answer.length for msg in self.all_known_messages if msg.answer}

    def weight_message(self, msg):
        return (msg[1].message.length not in self.known_sizes and msg[1].message.length or
                (1000 + ((msg[1].answer and msg[1].answer.length not in self.known_answer_sizes and
                          msg[1].answer.length) or 1000)))

    def order_messages(self):
        # first iteration: only unknown length messages
        for msg_index, msg_with_answer in sorted(enumerate(self.all_unknwown_messages), key=self.weight_message):
            msg = msg_with_answer.message
            new_item = (msg.length, msg_with_answer.answer and msg_with_answer.answer.length or 0)
            if new_item in self.already_tested:
                continue
            else:
                self.already_tested.add(new_item)
            yield msg_with_answer.message_id, msg

        # second iteration: non optimized, will do all so lot of redundancy
        for msg_index, msg_with_answer in sorted(enumerate(self.all_unknwown_messages), key=self.weight_message):
            msg = msg_with_answer.message
            yield msg_with_answer.message_id, msg
